/* nano-X launcher program
 * (C) 1999 Alistair Riddoch <ajr@ecs.soton.ac.uk>
 */

#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <stdlib.h>
#include <unistd.h>
#include "nano-X.h"



/*
 * Definitions to make it easy to define cursors
 */
#define	_	((unsigned) 0)		/* off bits */
#define	X	((unsigned) 1)		/* on bits */
#define	MASK(a,b,c,d,e,f,g) \
	(((((((((((((a * 2) + b) * 2) + c) * 2) + d) * 2) \
	+ e) * 2) + f) * 2) + g) << 9)

#define DEC_HEIGHT 7

static	GR_WINDOW_ID	w1;		/* id for large window */
static	GR_GC_ID	gc;		/* graphics context for rectangle */
static	GR_GC_ID	bgc;		/* graphics context for rectangle */
static	GR_SCREEN_INFO	si;		/* information about screen */
static	int		fwidth, fheight;
static	int		fbase;
static	int		num_apps = 0;

void do_exposure();
void do_launch();
void do_move();
void do_update();
void errorcatcher();			/* routine to handle errors */

struct app_info {
	char		app_id[10];
	char		app_path[64];
} Apps[] = {
#if ELKS
	{"clock", "/root/nclock"},
	{"term", "/root/nterm"},
	{"demo", "/root/demo"},
	{"demo2", "/root/demo2"},
#else
	{"clock", "bin/nclock"},
	{"term", "bin/nterm"},
	{"demo", "bin/demo"},
	{"demo2", "bin/demo2"},
#endif
	{"", ""}
};

typedef struct managed_window mwin;
struct managed_window {
	GR_WINDOW_ID	wid;
	GR_WINDOW_ID	fid;
	GR_COORD	x;
	GR_COORD	y;
	GR_SIZE		width;
	mwin		* next;
};

mwin * mwins = NULL;
mwin * in_motion = NULL;
GR_COORD	move_xoff;
GR_COORD	move_yoff;

int
main(int argc,char **argv)
{
	GR_EVENT	event;		/* current event */
	GR_BITMAP	bitmap1fg[7];	/* bitmaps for first cursor */
	GR_BITMAP	bitmap1bg[7];
	struct app_info	* act;
	int		width, height;

	for(act = Apps; act->app_id[0] != '\0'; act++, num_apps++);

	if (GrOpen() < 0) {
		fprintf(stderr, "cannot open graphics\n");
		exit(1);
	}
	
	GrGetScreenInfo(&si);

	GrSetErrorHandler(errorcatcher);

	gc = GrNewGC();
	bgc = GrNewGC();

	GrSetGCForeground(bgc, GRAY);
	GrSetGCFont(gc, GrCreateFont(FONT_OEM_FIXED, 0));

	GrGetGCTextSize(gc, "A", 1, TF_ASCII, &fwidth, &fheight, &fbase);
	width = fwidth * 8 + 4;
	height = (fheight) * num_apps + 4;

	w1 = GrNewWindow(GR_ROOT_WINDOW_ID, 5, 5, width,
		height, 1, WHITE, BLACK);

	GrSelectEvents(w1, GR_EVENT_MASK_EXPOSURE | GR_EVENT_MASK_BUTTON_DOWN);
	GrSelectEvents(GR_ROOT_WINDOW_ID, GR_EVENT_MASK_EXPOSURE | GR_EVENT_MASK_CHLD_UPDATE);

	GrMapWindow(w1);


	bitmap1bg[0] = MASK(_,_,X,X,X,_,_);
	bitmap1bg[1] = MASK(_,X,X,X,X,X,_);
	bitmap1bg[2] = MASK(_,X,X,X,X,X,_);
	bitmap1bg[3] = MASK(_,X,X,X,X,X,_);
	bitmap1bg[4] = MASK(_,X,X,X,X,X,_);
	bitmap1bg[5] = MASK(_,X,X,X,X,X,_);
	bitmap1bg[6] = MASK(X,X,X,X,X,X,X);

	bitmap1fg[0] = MASK(_,_,_,_,_,_,_);
	bitmap1fg[1] = MASK(_,_,_,X,_,_,_);
	bitmap1fg[2] = MASK(_,_,X,X,X,_,_);
	bitmap1fg[3] = MASK(_,_,X,X,X,_,_);
	bitmap1fg[4] = MASK(_,_,X,X,X,_,_);
	bitmap1fg[5] = MASK(_,_,X,X,X,_,_);
	bitmap1fg[6] = MASK(_,X,X,X,X,X,_);

	GrSetCursor(w1, 7, 7, 3, 3, WHITE, BLACK, bitmap1fg, bitmap1bg);

	GrFillRect(GR_ROOT_WINDOW_ID, bgc, 0, 0, si.cols, si.rows);

	GrSetGCForeground(gc, BLACK);
	GrSetGCBackground(gc, WHITE);

	while (1) {
		GrCheckNextEvent(&event);

		switch (event.type) {
			case GR_EVENT_TYPE_EXPOSURE:
				do_exposure(&event.exposure);
				break;
			case GR_EVENT_TYPE_BUTTON_DOWN:
				do_launch(&event.button);
				break;
			case GR_EVENT_TYPE_BUTTON_UP:
				do_move(&event.button);
				break;
			case GR_EVENT_TYPE_UPDATE:
				do_update(&event.update);
				break;
		}
	}
}

mwin * IsDecoration(GR_WINDOW_ID wid)
{
	mwin * mwp;
	for(mwp = mwins; mwp; mwp = mwp->next) {
		if (mwp->fid == wid) {
			return mwp;
		}
	}
	return NULL;
}
	

mwin * FindWindow(GR_WINDOW_ID wid)
{
	mwin * mwp;
	for(mwp = mwins; mwp; mwp = mwp->next) {
		if (mwp->wid == wid) {
			return mwp;
		}
	}
	return NULL;
}

mwin * NewWindow(GR_WINDOW_ID wid)
{
	mwin * mwp = malloc(sizeof(mwin));

	if (mwp) {
		mwp->wid = wid;
		mwp->next = mwins;
		mwins = mwp;
	}
	return mwp;
}

void
do_update(ep)
	GR_EVENT_UPDATE	*ep;
{
	mwin *	mwp;
	mwin *	tmwp;
	if (IsDecoration(ep->wid)) return;

	if ((mwp = FindWindow(ep->wid)) == NULL) {
		/* We have a new window */
		if (ep->utype != GR_UPDATE_MAP) return;
		printf("Decorating the new window\n");
		if ((mwp = NewWindow(ep->wid)) == NULL) {
			perror("malloc");
			return;
		}
		mwp->x = ep->x;
		mwp->y = ep->y;
		mwp->width = ep->width;
		GrMoveWindow(mwp->wid, mwp->x, mwp->y + DEC_HEIGHT);
		mwp->fid = GrNewWindow(GR_ROOT_WINDOW_ID, mwp->x, mwp->y,
				mwp->width, DEC_HEIGHT - 2, 1, BLUE, BLACK);
		GrSelectEvents(mwp->fid, GR_EVENT_MASK_BUTTON_DOWN | GR_EVENT_MASK_BUTTON_UP);
		GrMapWindow(mwp->fid);
	} else {
		switch (ep->utype) {
			case GR_UPDATE_UNMAP:
				GrUnmapWindow(mwp->fid);
				GrDestroyWindow(mwp->fid);
				if (mwins == mwp) {
					mwins = mwp->next;
				} else for(tmwp = mwins; tmwp; tmwp = tmwp->next) {
					if (tmwp->next == mwp) {
						tmwp->next = mwp->next;
					}
				}
				free(mwp);
				break;
			case GR_UPDATE_MOVE:
				if ((ep->x == mwp->x) && (ep->y == (mwp->y + DEC_HEIGHT))) {
					return;
				}
				mwp->x = ep->x;
				mwp->y = ep->y - DEC_HEIGHT;
				GrMoveWindow(mwp->fid, mwp->x, mwp->y);
			default:
				break;
		}
	}
}
/*
 * Here when an exposure event occurs.
 */
void
do_exposure(ep)
	GR_EVENT_EXPOSURE	*ep;
{
	struct app_info	* act;
	int app_no;

	if (ep->wid == w1) {
		for(act=Apps,app_no=0;act->app_id[0]!='\0';act++,app_no++) {
			GrText(w1, gc, 2, 2 - fbase + fheight * (app_no + 1),
				act->app_id, -1, 0);
		}
	} else if (ep->wid == GR_ROOT_WINDOW_ID) {
		GrFillRect(GR_ROOT_WINDOW_ID, bgc, ep->x, ep->y,
				ep->width, ep->height);
	}
}

extern char ** environ;

void
do_launch(ep)
	GR_EVENT_BUTTON	*ep;
{
	mwin *	mwp;
	static int app_no;

	if (ep->wid == w1) {
		app_no = ep->y / fheight;
		if (app_no >= num_apps) {
			app_no = num_apps - 1;
		}
		
		if (!vfork()) {
			char * nargv[2];

			nargv[0] = Apps[app_no].app_path;
			nargv[1] = 0;
			execve(nargv[0], nargv, environ);
			write(1, "\7", 1);
			exit(1);
		}
	} else if ((mwp = IsDecoration(ep->wid)) != NULL) {
		printf("Button down\n");
		GrRaiseWindow(mwp->wid);
		GrRaiseWindow(mwp->fid);
		in_motion = mwp;
		move_xoff = ep->x - mwp->x;
		move_yoff = ep->y - mwp->y;
		printf("{%d %d}\n", move_xoff, move_yoff);
 	}
}
 
void
do_move(ep)
GR_EVENT_BUTTON	*ep;
{
	mwin *	mwp;

	printf("Button up\n");
	if ((mwp = IsDecoration(ep->wid)) != NULL) {
		if (mwp == in_motion) {
			printf("move\n");
			mwp->x = ep->x - move_xoff;
			mwp->y = ep->y - move_yoff;
			GrMoveWindow(mwp->fid, mwp->x, mwp->y);
			GrMoveWindow(mwp->wid, mwp->x, mwp->y + DEC_HEIGHT);
			in_motion = NULL;
		}
	}
}

/*
 * Here on an unrecoverable error.
 */
void
errorcatcher(code, name, id)
	GR_ERROR	code;		/* error code */
	GR_FUNC_NAME	name;		/* function name which failed */
	GR_ID		id;		/* resource id */
{
	GrClose();
	fprintf(stderr, "DEMO ERROR: code %d, function %s, resource id %d\n",
		code, name, id);
	exit(1);
}
